/*
 * Copyright (C) 2019 Fabian Groffen <grobian@gentoo.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#ifndef _MUTT_ZSTRM_H_
#define _MUTT_ZSTRM_H_ 1

#include "mutt_socket.h"

#if defined(USE_ZLIB)
void mutt_zstrm_wrap_conn(CONNECTION *conn);
#endif

#endif /* _MUTT_ZSTRM_H_ */
